/*

  Compiler settings for MSVC 4.2b:

  (suggested)

  preprocessory defs:  NDEBUG ( not  _DEBUG )
  calling convention: fastcall
  processor: pentium
 
   optimizations:
      Assume No Aliasing
      Global Optimizations 
      Generate Intrinsic Functions
      Favor Fast Code 
      Frame-Pointer Omission 
      Full Optimization

  In-line Function Expansion
    Only __inline

  not using precompiled headers
  not incrementally linked or compiled

*/

#include "first.h"
#include "raster_util.h"
#include "clut.h"
#include "raster.h"
#include "rast_i.h"
#include <assert.h>

#pragma optimize("a", on)

using namespace raster;


//const float min_shade =  1.0f/64.0f;
//const float max_shade =  1.0f - 1.0f/128.0f;


//
// number if fixed point bits to use for u and v
//
const int uv_fx_bits = 20;
const int shade_fx_bits = 16;


#define FSHL(exp)  ( (127+( exp )) << 23)

const int __shade_add = 0;
const int i_p_min_shade = FSHL(23-shade_fx_bits + (__shade_add)) + (256*1  >>(__shade_add));
const int i_p_max_shade = FSHL(23-shade_fx_bits + (__shade_add)) + (256*255>>(__shade_add));





//
// size of blocks for bilinear interpolation
// 4 means 16,16 blocks higher values are faster but less perspectively 
// correct
//
// note: these values must be synched with rast_asm.asm
//
const int min_block_bits = 2;
const int min_block_dim  = (1<<min_block_bits);


const int max_x_blocks =  max_x_res/min_block_dim + 1;
const int max_y_blocks =  max_y_res/min_block_dim + 1;


//
//  data stored at each point in grid
//

struct pt_dat
{
 int shade;
 int u, v;
};


//
// pt dats at last y
//
static pt_dat       last_y_dats[max_x_blocks];

// block extents 
static byte block_extents[max_y_blocks*2];




//
// packed data created from pt_dats, and interpolated along y 
//

struct block_data
{
 __int64 uvs,d_dy_uvs;
};

block_data p_block_datas[max_x_blocks];






float plt_u_magic;
float plt_shade_magic;
// depends on size of texture
float plt_v_magic;


inline void scale_flt(float & f, int power )
{
 (int&)f += power << 23;
}



template <int bpp, int block_bits, int has_transparency>
struct p_drawer
{
 static inline int block_dim() { return (1<<block_bits); }

  //
 // rounds down
 static inline int get_block0( int x0 ) { return x0 >> block_bits; }

 // rounds up
 static inline int get_block1( int x1 ) { return (x1+block_dim()-1) >> block_bits; }

 //
//  forward difference between 2 points
//
 static inline int fdif( int end, int start) { return (end-start)>>block_bits; }

 static inline void set_magics(const texture * tmap)
 {
  plt_v_magic       =  fshl_hf( 23 - uv_fx_bits +block_bits + 8  - tmap->l2_dim[1] );
  plt_u_magic       =  fshl_hf( 23 - uv_fx_bits +block_bits  );
  plt_shade_magic   =  fshl( 23 - shade_fx_bits +block_bits*0  );
 }

 static void __fastcall set_asm_code(const texture * tmap );
 static void __fastcall draw_block( int y );
 
 static void draw(const texture * tmap);

 static void make_block_data(int y, int x0, int x1);
 
 static const texture * last_tex;
};

// flat shaded
template <int bpp, int block_bits, int has_transparency>
struct p_drawer_f : p_drawer<bpp,block_bits,has_transparency>
{
 static void draw(const texture * tmap);
 static void make_block_data(int y, int x0, int x1);
};





template <int bpp, int block_bits >
struct p_drawer_s
{
 static inline int block_dim() { return (1<<block_bits); }

  //
 // rounds down
 static inline int get_block0( int x0 ) { return x0 >> block_bits; }

 // rounds up
 static inline int get_block1( int x1 ) { return (x1+block_dim()-1) >> block_bits; }

 //
//  forward difference between 2 points
//
 static inline int fdif( int end, int start) { return (end-start)>>block_bits; }

 static inline void set_magics(const solid_surface_desc * sd )
 {
  plt_shade_magic   =  fshl( 23 - (8 + sd->shade_bits )  + block_bits*0  );
 }

 static void __fastcall draw_block( int y );
 
 static void draw(const solid_surface_desc * sd);

 static void make_block_data(int y, int x0, int x1);
 
};



typedef  void span_draw_func(const texture * tmap);
typedef  void span_draw_func_s(const solid_surface_desc * sd);


static  span_draw_func * all_funcs[] = { p_drawer<8,2,0>::draw, 
                                         p_drawer<8,2,1>::draw, 
                                         p_drawer<8,3,0>::draw, 
                                         p_drawer<8,3,1>::draw, 
                                         p_drawer<8,4,0>::draw,
                                         p_drawer<8,4,1>::draw,
                                         p_drawer<16,2,0>::draw, 
                                         p_drawer<16,2,1>::draw, 
                                         p_drawer<16,3,0>::draw, 
                                         p_drawer<16,3,1>::draw, 
                                         p_drawer<16,4,0>::draw,
                                         p_drawer<16,4,1>::draw,
                                           };

static  span_draw_func * all_funcs_f[] = { p_drawer_f<8,2,0>::draw, 
                                            p_drawer_f<8,2,1>::draw, 
                                         p_drawer_f<8,3,0>::draw, 
                                         p_drawer_f<8,3,1>::draw, 
                                         p_drawer_f<8,4,0>::draw,
                                         p_drawer_f<8,4,1>::draw,
                                         p_drawer_f<16,2,0>::draw, 
                                         p_drawer_f<16,2,1>::draw, 
                                         p_drawer_f<16,3,0>::draw, 
                                         p_drawer_f<16,3,1>::draw, 
                                         p_drawer_f<16,4,0>::draw,
                                         p_drawer_f<16,4,1>::draw,
                                           };

static  span_draw_func_s * all_funcs_s[] = {
                                         p_drawer_s<8,2>::draw,
                                         p_drawer_s<8,3>::draw, 
                                         p_drawer_s<8,4>::draw,
                                         p_drawer_s<16,2>::draw, 
                                         p_drawer_s<16,3>::draw, 
                                         p_drawer_s<16,4>::draw,
                                           };


static  span_draw_func ** cur_draw_func;
static  span_draw_func ** cur_draw_func_f;

static  span_draw_func_s * cur_draw_func_s;


void raster::refresh_persp_draw_funcs()
{
 assert( (block_bits >= 2)&&(block_bits <= 4) );

 cur_draw_func     =  &all_funcs[ (block_bits - 2  + ((bytes_per_pixel-1)*3))*2   ];
 cur_draw_func_f   =  &all_funcs_f[ (block_bits - 2  + ((bytes_per_pixel-1)*3))*2   ];
 
 cur_draw_func_s =  all_funcs_s[ (block_bits - 2  + ((bytes_per_pixel-1)*3))   ];

}






//
// pack u, v, shade into 1 64 bit number  
// u: 32 - v_bits
// v: v_bits + 16
// shade: 16
//
// this used to be an inline, but msvc won't inline it
//
#define  pack_uvshade(rv, u, v, shade)                                             \
{                                                                                  \
    ((int*)&rv)[1]  =  (u<< (32 -uv_fx_bits) ) | ( (v>> (uv_fx_bits-8) )&0xff );   \
    (int&)rv        =  (v<< (40 -uv_fx_bits) ) | ( (shade ) &0xfffff);              \
}

#define  pack_shade(rv,shade)                                               \
{                                                                           \
    (int&)rv = shade&0xfffff;                                                \
}







template <int bpp, int block_bits, int has_transparency>
void p_drawer<bpp,block_bits,has_transparency>::make_block_data(int y, int x0, int x1)
{
                
 pt_dat       p1; // current point that we are making
 
 float zb_ = plb_z[0] + plb_z[2]*y + plb_z[1]*x0;
 float ub_ = plb_u[0] + plb_u[2]*y + plb_u[1]*x0;
 float vb_ = plb_v[0] + plb_v[2]*y + plb_v[1]*x0;
 float sb_ = plb_sh[0]+ plb_sh[2]*y+ plb_sh[1]*x0;


 float z = 1.0f/zb_;

 do
 { 

     (float&)p1.u     = z * ub_ + plt_u_magic;
     (float&)p1.v     = z * vb_ + plt_v_magic;
     
     // 
     //  need to clamp shade
     // 
     float tshade = z * sb_ * (1.0 -  z * raster::haze_far_z_inv  );

/*     if( (int&)tshade < (int&)min_shade)
       tshade = min_shade;
     if( (int&)tshade > (int&)max_shade)
       tshade = max_shade;
*/

     (float&)p1.shade = tshade + plt_shade_magic;

	 p1.shade &= ~( (1<<block_bits) - 1);
 
	 if(  p1.shade >  i_p_max_shade)
	    p1.shade =  i_p_max_shade;
	 
	 if(  p1.shade <  i_p_min_shade)
	    p1.shade =  i_p_min_shade;

     // update by d_dx
     ub_ += plb_u[1];
     vb_ += plb_v[1];
     sb_ += plb_sh[1];
     zb_ += plb_z[1];
     
      // make next z
   
     z = 1.0f/zb_;

   // is this the first block?
     pt_dat & p0 = last_y_dats[x0];

     // make new value
     // are we could be going back too far?

      block_data & bd = p_block_datas[x0];
    
      pack_uvshade(bd.uvs, p0.u,p0.v,p0.shade);
    
      pack_uvshade(bd.d_dy_uvs, fdif(p1.u,p0.u),
                                  fdif(p1.v,p0.v),
                                  fdif(p1.shade,p0.shade)  );
         // shift down by 1
     // p0 = p1;
       p0.u = p1.u;
       p0.v = p1.v;
       p0.shade = p1.shade;




  } while( ++x0, x0 <= x1 );
}


template <int bpp, int block_bits, int has_transparency>
void p_drawer_f<bpp,block_bits,has_transparency>::make_block_data(int y, int x0, int x1)
{
                
 pt_dat       p1; // current point that we are making
 
 float zb_ = plb_z[0] + plb_z[2]*y + plb_z[1]*x0;
 float ub_ = plb_u[0] + plb_u[2]*y + plb_u[1]*x0;
 float vb_ = plb_v[0] + plb_v[2]*y + plb_v[1]*x0;


 float shade = plb_sh[0] + plt_shade_magic;
 int ishade = (int&)shade;

 float z = 1.0f/zb_;

 do
 { 

     (float&)p1.u     = z * ub_ + plt_u_magic;
     (float&)p1.v     = z * vb_ + plt_v_magic;
     

     // update by d_dx
     ub_ += plb_u[1];
     vb_ += plb_v[1];
     zb_ += plb_z[1];
     
      // make next z
   
     z = 1.0f/zb_;

   // is this the first block?
     pt_dat & p0 = last_y_dats[x0];

     // make new value
     // are we could be going back too far?

      block_data & bd = p_block_datas[x0];
    
      pack_uvshade(bd.uvs, p0.u,p0.v,ishade);
    
      pack_uvshade(bd.d_dy_uvs, fdif(p1.u,p0.u),
                                  fdif(p1.v,p0.v),
                                  0  );
         // shift down by 1
       p0.u = p1.u;
       p0.v = p1.v;

  } while( ++x0, x0 <= x1 );
}



template <int bpp, int block_bits>
void p_drawer_s<bpp,block_bits>::make_block_data(int y, int x0, int x1)
{
              
 pt_dat       p1; // current point that we are making
 
 float zb_ = plb_z[0] + plb_z[2]*y + plb_z[1]*x0;
 float sb_ = plb_sh[0]+ plb_sh[2]*y+ plb_sh[1]*x0;

 float z = 1.0f/zb_;

 do
 { 
    
     // 
     //  need to clamp z. this is not very efficient
     // 
     float tshade = z * sb_ * (1.0 -  z * raster::haze_far_z_inv  );

   /*  if( (int&)tshade < (int&)min_shade)
       tshade = min_shade;
     if( (int&)tshade > (int&)max_shade)
       tshade = max_shade;

*/
     (float&)p1.shade = tshade + plt_shade_magic;


	 if(  p1.shade >  i_p_max_shade)
       p1.shade =  i_p_max_shade;
	 
	 if(  p1.shade <  i_p_min_shade)
	    p1.shade =  i_p_min_shade;

  
     // update by d_dx
     sb_ += plb_sh[1];
     zb_ += plb_z[1];
     
      // make next z
   
     z = 1.0f/zb_;

   // is this the first block?
     pt_dat & p0 = last_y_dats[x0];

     // make new value
     // are we could be going back too far?

      block_data & bd = p_block_datas[x0];
    
      pack_shade(bd.uvs, p0.shade);
    
      pack_shade(bd.d_dy_uvs,    fdif(p1.shade,p0.shade)  );
      // shift down by 1
     p0.shade = p1.shade;

  } while( ++x0, x0 <= x1 );
}



template <int bpp, int block_bits, int has_transparency>
const texture  * p_drawer<bpp,block_bits,has_transparency>::last_tex;


//
// Minimum and maximum macros (same as in stdlib) 
//
#ifndef __max
#define __max(a,b)  (((a) > (b)) ? (a) : (b))
#define __min(a,b)  (((a) < (b)) ? (a) : (b))
#endif


template <int bpp, int block_bits, int has_transparency>
void p_drawer<bpp,block_bits,has_transparency>::draw(const texture * tmap)
{

  //
  // find block extents
  //
  
  {
      int y =  y_min;
      byte * block_ptr = block_extents;
      
      // last min/max
      int l_min_x = max_x_res, l_max_x=0;
      
      while( y != y_max )
      {

       int min_x = span_extents[y][0];
       int max_x = span_extents[y][1];
      
       int y_end = __min(y_max,y+block_dim());
    
       // pre inc
       while(++y < y_end)
       {
         int x0 =span_extents[y][0];
         int x1 =span_extents[y][1];
         
         if( x0 < min_x ) min_x = x0;
         if( x1 > max_x ) max_x = x1;
       }
    
       block_ptr[0] =  get_block0(__min(min_x,l_min_x));
       block_ptr[1] =  get_block1(__max(max_x,l_max_x));
       
       block_ptr += 2;

       l_min_x = min_x;
       l_max_x = max_x;
      }  
      // set extra one
      block_ptr[0] =  get_block0(l_min_x);
      block_ptr[1] =  get_block1(l_max_x);
 }

  //
  // render blocks
  //

    set_magics(tmap);

    if( last_tex != tmap ) 
    {
      last_tex = tmap;
 
      set_asm_code( tmap );

    }


 //
 // scale d_dx's by block_dim
 //
  scale_flt( plb_sh[1], block_bits  );
  scale_flt( plb_u[1] , block_bits  );
  scale_flt( plb_v[1] , block_bits  );
  scale_flt( plb_z[1] , block_bits  );
 

  // make first block
  make_block_data( y_min,block_extents[0],block_extents[1]);

  byte * block_ptr = &block_extents[2];
  int cur_y = y_min;
  do
  {
   // make next line
    
    int eb0 =  block_ptr[0];
    int eb1 =  block_ptr[1];
    block_ptr += 2;
    
    int end_y = cur_y + block_dim();
    make_block_data(end_y,eb0,eb1 );
    if( end_y > y_max) 
     end_y = y_max;

   p_start_block = &p_block_datas[eb0];
   p_end_block   = &p_block_datas[eb1];

   
    for(;;)
    {
     draw_block(cur_y);
          
     ++cur_y;
     if( cur_y >= end_y)
       break;
      
    } // for ys
   

  } while( cur_y != y_max ); // for each y block
}


template <int bpp, int block_bits, int has_transparency>
void p_drawer_f<bpp,block_bits,has_transparency>::draw(const texture * tmap)
{

  //
  // find block extents
  //
  
  {
      int y =  y_min;
      byte * block_ptr = block_extents;
      
      // last min/max
      int l_min_x = max_x_res, l_max_x=0;
      
      while( y != y_max )
      {

       int min_x = span_extents[y][0];
       int max_x = span_extents[y][1];
      
       int y_end = __min(y_max,y+block_dim());
    
       // pre inc
       while(++y < y_end)
       {
         int x0 =span_extents[y][0];
         int x1 =span_extents[y][1];
         
         if( x0 < min_x ) min_x = x0;
         if( x1 > max_x ) max_x = x1;
       }
    
       block_ptr[0] =  get_block0(__min(min_x,l_min_x));
       block_ptr[1] =  get_block1(__max(max_x,l_max_x));
       
       block_ptr += 2;

       l_min_x = min_x;
       l_max_x = max_x;
      }  
      // set extra one
      block_ptr[0] =  get_block0(l_min_x);
      block_ptr[1] =  get_block1(l_max_x);
 }

  //
  // render blocks
  //

    set_magics(tmap);

    if( last_tex != tmap ) 
    {
      last_tex = tmap;
 
      set_asm_code( tmap );

    }


 //
 // scale d_dx's by block_dim
 //
  scale_flt( plb_u[1] , block_bits  );
  scale_flt( plb_v[1] , block_bits  );
  scale_flt( plb_z[1] , block_bits  );
 

  // make first block
  make_block_data( y_min,block_extents[0],block_extents[1]);

  byte * block_ptr = &block_extents[2];
  int cur_y = y_min;
  do
  {
   // make next line
    
    int eb0 =  block_ptr[0];
    int eb1 =  block_ptr[1];
    block_ptr += 2;
    
    int end_y = cur_y + block_dim();
    make_block_data(end_y,eb0,eb1 );
    if( end_y > y_max) 
     end_y = y_max;

   p_start_block = &p_block_datas[eb0];
   p_end_block   = &p_block_datas[eb1];

   
    for(;;)
    {
     draw_block(cur_y);
          
     ++cur_y;
     if( cur_y >= end_y)
       break;
      
    } // for ys
   

  } while( cur_y != y_max ); // for each y block
}




extern "C" void * cur_solid_clut;

template <int bpp, int block_bits>
void p_drawer_s<bpp,block_bits>::draw(const solid_surface_desc * sd)
{
 cur_solid_clut = sd->clut;
 
  //
  // find block extents
  //
  
  {
      int y =  y_min;
      byte * block_ptr = block_extents;
      
      // last min/max
      int l_min_x = max_x_res, l_max_x=0;
      
      while( y != y_max )
      {

       int min_x = span_extents[y][0];
       int max_x = span_extents[y][1];
      
       int y_end = __min(y_max,y+block_dim());
    
       // pre inc
       while(++y < y_end)
       {
         int x0 =span_extents[y][0];
         int x1 =span_extents[y][1];
         
         if( x0 < min_x ) min_x = x0;
         if( x1 > max_x ) max_x = x1;
       }
    
       block_ptr[0] =  get_block0(__min(min_x,l_min_x));
       block_ptr[1] =  get_block1(__max(max_x,l_max_x));
       
       block_ptr += 2;

       l_min_x = min_x;
       l_max_x = max_x;
      }  
      // set extra one
      block_ptr[0] =  get_block0(l_min_x);
      block_ptr[1] =  get_block1(l_max_x);
 }

  //
  // render blocks
  //

    set_magics(sd);


 //
 // scale d_dx's by block_dim
 //
  scale_flt( plb_sh[1], block_bits  );
  scale_flt( plb_z[1] , block_bits  );
 

  // make first block
  make_block_data( y_min,block_extents[0],block_extents[1]);

  byte * block_ptr = &block_extents[2];
  int cur_y = y_min;
  do
  {
   // make next line
    
    int eb0 =  block_ptr[0];
    int eb1 =  block_ptr[1];
    block_ptr += 2;
    
    int end_y = cur_y + block_dim();
    make_block_data(end_y,eb0,eb1 );
    if( end_y > y_max) 
     end_y = y_max;

   p_start_block = &p_block_datas[eb0];
   p_end_block   = &p_block_datas[eb1];

   
    for(;;)
    {
     draw_block(cur_y);
          
     ++cur_y;
     if( cur_y >= end_y)
       break;
      
    } // for ys
   

  } while( cur_y != y_max ); // for each y block
}








block_data * p_start_block;
block_data * p_end_block;





                          










//
// perspectively decent
//
void raster::draw_poly(const poly_plt * p)
{
 // get first 3 verts
  const vert_2d * v0 = p->verts;
  const vert_2d * v1 = v0->next;
  const vert_2d * v2 = v1->next;


  float x0 = v0->x;
  float y0 = v0->y;

  float dx1 = (v1->x - x0);
  float dy1 = (v1->y - y0);

  float dx2 = (v2->x - x0);
  float dy2 = (v2->y - y0);

  float area = dx2*dy1 - dy2*dx1;
  
  //
  // is poly too small to draw, or back facing?
  // note: if verts are entered in wrong order, e.g.
  // cw insetad of ccw, they will not be drawn. (2d backface culling)
  
  if( area <= min_tri_area  )
  {
     return;
  } 
  
  //
  // trace out edges
  //
  
  INIT_FP
  
  reset_spans();
  const vert_2d* e0 = v0;
  do
   {
    const vert_2d* e1 = e0->next;

    trace_edge_func( e0->x,e0->y,e1->x,e1->y);
    e0 = e1;
  } while( e0 != v0 );
  
  
  //
  // put fp unit into single precision mode
  //
  set_single_pres();


  // doesn't cover any vertical space?
  if( y_min >= y_max )
  {
   RESTORE_FP_FROM_SINGLE
   return;
  }

 

  //
  // do a cramer inverse on the 2d vertex system
  //
  float det_  =  -1.0/area;

  float  xf_mat00 =  dy2*det_;
  float  xf_mat01 = -dx2*det_;
  float  xf_mat10 = -dy1*det_;
  float  xf_mat11 =  dx1*det_;

 
   const float * z_s = p->z_invs;

   LB_TO_SCREEN_P(plb_sh,p->shades,x0,y0)
   LB_TO_SCREEN_P(plb_u ,p->us    ,x0,y0)
   LB_TO_SCREEN_P(plb_v ,p->vs    ,x0,y0)
   LB_TO_SCREEN(plb_z ,z_s,x0,y0)

   cur_draw_func[p->tmap->has_transparency]( p->tmap );



  RESTORE_FP_FROM_SINGLE
}






void raster::draw_poly(const poly_pft * p)
{
 // get first 3 verts
  const vert_2d* v0 = p->verts;
  const vert_2d* v1 = v0->next;
  const vert_2d* v2 = v1->next;


  float x0 = v0->x;
  float y0 = v0->y;

  float dx1 = (v1->x - x0);
  float dy1 = (v1->y - y0);

  float dx2 = (v2->x - x0);
  float dy2 = (v2->y - y0);

  float area = dx2*dy1 - dy2*dx1;
  
  //
  // is poly too small to draw, or back facing?
  // note: if verts are entered in wrong order, e.g.
  // cw insetad of ccw, they will not be drawn. (2d backface culling)
  
  if( area <= min_tri_area  )
  {
     return;
  } 
  
  //
  // trace out edges
  //
  
  INIT_FP
  
  reset_spans();
  const vert_2d* e0 = v0;
  do
   {
    const vert_2d* e1 = e0->next;

    trace_edge_func( e0->x,e0->y,e1->x,e1->y);
    e0 = e1;
  } while( e0 != v0 );
  
  
  //
  // put fp unit into single precision mode
  //
  set_single_pres();


  // doesn't cover any vertical space?
  if( y_min >= y_max )
  {
   RESTORE_FP_FROM_SINGLE
   return;
  }

 

  //
  // do a cramer inverse on the 2d vertex system
  //
  float det_  =  -1.0/area;

  float  xf_mat00 =  dy2*det_;
  float  xf_mat01 = -dx2*det_;
  float  xf_mat10 = -dy1*det_;
  float  xf_mat11 =  dx1*det_;

 
   const float * z_s = p->z_invs;

   plb_sh[0] = p->shade*linear_shade_scale +  linear_shade_eps;
   LB_TO_SCREEN_P(plb_u ,p->us    ,x0,y0)
   LB_TO_SCREEN_P(plb_v ,p->vs    ,x0,y0)
   LB_TO_SCREEN(plb_z ,z_s,x0,y0)

   cur_draw_func_f[p->tmap->has_transparency]( p->tmap );



  RESTORE_FP_FROM_SINGLE
}

//
// solid
//
void raster::draw_poly(const poly_pls * p)
 {
 // get first 3 verts
  const vert_2d* v0 = p->verts;
  const vert_2d* v1 = v0->next;
  const vert_2d* v2 = v1->next;


  float x0 = v0->x;
  float y0 = v0->y;

  float dx1 = (v1->x - x0);
  float dy1 = (v1->y - y0);

  float dx2 = (v2->x - x0);
  float dy2 = (v2->y - y0);

  float area = dx2*dy1 - dy2*dx1;
  
  //
  // is poly too small to draw, or back facing?
  // note: if verts are entered in wrong order, e.g.
  // cw insetad of ccw, they will not be drawn. (2d backface culling)
  
  if( area <= min_tri_area  )
  {
     return;
  } 
  
  //
  // trace out edges
  //
  
  INIT_FP
  
  reset_spans();
  const vert_2d* e0 = v0;
  do
   {
    const vert_2d* e1 = e0->next;

    trace_edge_func( e0->x,e0->y,e1->x,e1->y);
    e0 = e1;
  } while( e0 != v0 );
  
  
  //
  // put fp unit into single precision mode
  //
  set_single_pres();


  // doesn't cover any vertical space?
  if( y_min >= y_max )
  {
   RESTORE_FP_FROM_SINGLE
   return;
  }

 

  //
  // do a cramer inverse on the 2d vertex system
  //
  float det_  =  -1.0/area;

  float  xf_mat00 =  dy2*det_;
  float  xf_mat01 = -dx2*det_;
  float  xf_mat10 = -dy1*det_;
  float  xf_mat11 =  dx1*det_;

 
   const float * z_s = p->z_invs;

   LB_TO_SCREEN_P(plb_sh,p->shades,x0,y0)
   LB_TO_SCREEN(plb_z ,z_s,x0,y0)

   cur_draw_func_s( p->surf );



  RESTORE_FP_FROM_SINGLE
}



/*

static U16 tmp_buf[768][2];
extern U16  * frame_buf_starts[max_y_res];
void __save_state()
{
   for(int y=y_min;y<y_max;++y)
   {
    tmp_buf[y][0] = frame_buf_starts[y][span_extents[y][0]-1];
    tmp_buf[y][1] = frame_buf_starts[y][span_extents[y][1]];
   }
}
void __assert_state()
{
   for(int y=y_min;y<y_max;++y)
   {
    assert( tmp_buf[y][0] == frame_buf_starts[y][span_extents[y][0]-1]);
    assert( tmp_buf[y][1] == frame_buf_starts[y][span_extents[y][1]]);
   }
}
*/ 
